## Copyright (C) 2012 Rik Wehbring
## Copyright (C) 1995-2016 Kurt Hornik
## Copyright (C) 2023 Andreas Bertsatos <abertsatos@biol.uoa.gr>
##
## This program is free software: you can redistribute it and/or
## modify it under the terms of the GNU General Public License as
## published by the Free Software Foundation, either version 3 of the
## License, or (at your option) any later version.
##
## This program is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; see the file COPYING.  If not, see
## <http://www.gnu.org/licenses/>.

## -*- texinfo -*-
## @deftypefn  {statistics} @var{y} = geopdf (@var{x}, @var{ps})
##
## Geometric probability density function (PDF).
##
## For each element of @var{x}, compute the probability density function (PDF)
## at @var{x} of the geometric distribution with parameter @var{ps}.  The size
## of @var{y} is the common size of @var{x} and @var{ps}.  A scalar input
## functions as a constant matrix of the same size as the other inputs.
##
## The geometric distribution models the number of failures (@var{x}) of a
## Bernoulli trial with probability @var{ps} before the first success.
##
## @seealso{geocdf, geoinv, geornd, geostat}
## @end deftypefn

function y = geopdf (x, ps)

  if (nargin != 2)
    print_usage ();
  endif

  if (! isscalar (ps))
    [retval, x, ps] = common_size (x, ps);
    if (retval > 0)
      error ("geopdf: X and PS must be of common size or scalars.");
    endif
  endif

  if (iscomplex (x) || iscomplex (ps))
    error ("geopdf: X and PS must not be complex.");
  endif

  if (isa (x, "single") || isa (ps, "single"))
    y = zeros (size (x), "single");
  else
    y = zeros (size (x));
  endif

  k = isnan (x) | (x == Inf) | !(ps >= 0) | !(ps <= 1);
  y(k) = NaN;

  k = (x >= 0) & (x < Inf) & (x == fix (x)) & (ps > 0) & (ps <= 1);
  if (isscalar (ps))
    y(k) = ps * ((1 - ps) .^ x(k));
  else
    y(k) = ps(k) .* ((1 - ps(k)) .^ x(k));
  endif

endfunction


%!shared x,y
%! x = [-1 0 1 Inf];
%! y = [0, 1/2, 1/4, NaN];
%!assert (geopdf (x, 0.5*ones (1,4)), y)
%!assert (geopdf (x, 0.5), y)
%!assert (geopdf (x, 0.5*[-1 NaN 4 1]), [NaN NaN NaN y(4)])
%!assert (geopdf ([x, NaN], 0.5), [y, NaN])

## Test class of input preserved
%!assert (geopdf (single ([x, NaN]), 0.5), single ([y, NaN]), 5*eps ("single"))
%!assert (geopdf ([x, NaN], single (0.5)), single ([y, NaN]), 5*eps ("single"))

## Test input validation
%!error geopdf ()
%!error geopdf (1)
%!error geopdf (1,2,3)
%!error geopdf (ones (3), ones (2))
%!error geopdf (ones (2), ones (3))
%!error geopdf (i, 2)
%!error geopdf (2, i)
